#!/usr/local/bin/python3.10

import os 
import time
dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)
print(dir_path)

# minikube kubectl -- create -f ../start.yaml ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip

import requests

import tkinter as tk
import tkinter.messagebox
import customtkinter
from customtkinter import ThemeManager
import json
import os
from PIL import Image, ImageTk
import traceback
from multiprocessing import Process
from multiprocessing import Queue

import re
import pandas as pd
import numpy as np
import ast
import platform
import time

from queue import Empty

from .General.NoticeWindow import NoticeWindow as NW

from ..util import PSORunner
from ..util import GraphGenerator
from ..util.CTkToolTip import CTkToolTip as ctt
from ..util import helpers as hp

import subprocess
import plotly.express as px
import plotly.graph_objs as go

from tkinter.filedialog import askopenfilename
from tkinter.filedialog import asksaveasfilename

from . import OptionManager as om

from .SetupTab import SetupTab as st
from .PlatformTab import PlatformTab as pt
from .RunTab import RunTab as rt
from .VisualizeTab import VisualizeTab as vt

from ..util.CTkToolTip import CTkToolTip as ctt

customtkinter.set_appearance_mode("System")  # Modes: "System" (standard), "Dark", "Light"
customtkinter.set_default_color_theme("blue")  # Themes: "blue" (standard), "green", "dark-blue"

results_queue = Queue()
stdout_queue = Queue()
stderr_queue = Queue()

class App(customtkinter.CTk):
	def __init__(self):
		super().__init__()
		
		self.option_manager = om.OptionManager()
		
		# self.graph_selector_value = tk.StringVar()
		# self.graph_selector_value.set("Best Cost Stacked")

		# self.graph_theme_value = tk.StringVar()
		# self.graph_theme_value.set("Dark")

		# self.selected_csv = tk.StringVar()
		# self.selected_csv.set("No files found...")
		self.open_file = "None"
		self.csv_data = None

		# self.selected_csv2 = tk.StringVar()
		# self.selected_csv2.set("No files found...")
		self.open_file2 = "None"
		self.csv_data2 = None

		# self.selected_x = tk.StringVar()
		# self.selected_x.set("time")

		# self.selected_y1 = tk.StringVar()
		# self.selected_y1.set("NONE")

		# self.selected_y2 = tk.StringVar()
		# self.selected_y2.set("NONE")

		# self.figure_style = tk.StringVar()
		# self.figure_style.set("Scatter")

		# self.matrix_values = []
		# self.matrix_values.append(tk.StringVar())
		# self.matrix_values[0].set("NONE")

		self.running_config = None
		self.selected_graph_name = None

		self.train_process = None
		self.minikube_process = None
		self.data_x = [0]
		self.data_y = [0]
		
		self.image_scale = 1
		self.image_width = 1280
		self.image_height = 720
		self.progress_data = None
		self.calibration_data = None
		self.testing = False

		# configure window
		self.title("COSU Manager")
		self.geometry(f"{1920}x{1080}")

		# configure grid layout (4x4)
		self.grid_columnconfigure(0, weight=1)
		self.grid_rowconfigure(1, weight=1)

		header_padding_x = (5, 5)
		header_padding_y = (10, 10)
		
		self.sidebar_frame = customtkinter.CTkFrame(self, width=140, corner_radius=0)
		self.sidebar_frame.grid(row=0, column=0, sticky="nsew")
		self.sidebar_frame.grid_columnconfigure(4, weight=1)
		self.logo_label = customtkinter.CTkLabel(self.sidebar_frame, text="COSU Manager", font=customtkinter.CTkFont(size=20, weight="bold"))
		self.logo_label.grid(row=0, column=0, padx=(20, 10), pady=header_padding_y)
		self.save_button = customtkinter.CTkButton(self.sidebar_frame, text="Save", width=60, command=self.save_project)
		self.save_button.grid(row=0, column=1, padx=header_padding_x, pady=header_padding_y)
		self.load_button = customtkinter.CTkButton(self.sidebar_frame, text="Load", width=60, command=self.load_project)
		self.load_button.grid(row=0, column=2, padx=header_padding_x, pady=header_padding_y)
		
		# 4 is service URL
		self.service_label = customtkinter.CTkLabel(self.sidebar_frame, text="Service:", anchor="w")
		self.service_label.grid(row=0, column=3, padx=(80, 5), pady=header_padding_y)
		self.service_url = customtkinter.CTkEntry(self.sidebar_frame, textvariable=self.option_manager.get("url"))
		self.service_url.grid(row=0, column=4, columnspan=1, padx=header_padding_x, pady=header_padding_y, sticky="nsew")
		refresh_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "refresh.png")), size=(20, 20))
		self.refresh_button = customtkinter.CTkButton(self.sidebar_frame, text=None, width=30, image=refresh_image, command=self.load)
		ctt(self.refresh_button, delay=0.1, alpha=0.95, message="Connect to Service")
		self.refresh_button.grid(row=0, column=5, padx=header_padding_x, pady=header_padding_y)

		self.algorithm_optionmenu = customtkinter.CTkOptionMenu(self.sidebar_frame, variable=self.option_manager.get_mode_sv(), values=self.option_manager.get_service_modes(), width=50, command=self.refresh_step_view)
		self.algorithm_optionmenu.grid(row=0, column=6, padx=(5, 80), pady=header_padding_y)
		
		self.scaling_label = customtkinter.CTkLabel(self.sidebar_frame, text="Scale:", anchor="w")
		self.scaling_label.grid(row=0, column=7, padx=header_padding_x, pady=header_padding_y)
		self.scaling_optionemenu = customtkinter.CTkOptionMenu(self.sidebar_frame, values=["50%", "75%", "100%", "125%", "150%", "175%", "200%"], width=60,
																command=self.change_scaling_event)
		self.scaling_optionemenu.grid(row=0, column=8, padx=(5, 20), pady=header_padding_y)
		self.scaling_optionemenu.set("100%")
		
		expand_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "expand.png")), size=(20, 20))
		#self.new_window = customtkinter.CTkButton(self.sidebar_frame, text=None, width=30, image=expand_image, command=self.open_new_window)
		#ctt(self.new_window, delay=0.1, alpha=0.95, message="Open New Window")
		#self.new_window.grid(row=0, column=9, padx=(5, 20), pady=header_padding_y)
		
		self.tabview = customtkinter.CTkTabview(self, bg_color="transparent", fg_color="transparent")
		self.tabview.grid(row=1, column=0, padx=(0, 0), pady=(10, 10), sticky="nsew")
		tab1 = "Platform"
		tab2 = "Setup"
		tab3 = "Project"
		tab4 = "Results"
		
		self.tabview.add(tab1)
		self.tabview.add(tab2)
		self.tabview.add(tab4)
		self.tabview.add(tab3)
		
		pt.create_tab(self, self.tabview.tab(tab1))
		st.create_tab(self, self.tabview.tab(tab2))
		rt.create_tab(self, self.tabview.tab(tab3))
		vt.create_tab(self, self.tabview.tab(tab4))
		
		self.footer_frame = customtkinter.CTkFrame(self, corner_radius=0)
		self.footer_frame.grid(row=2, column=0, sticky="nsew")
		self.footer_frame.grid_columnconfigure(4, weight=1)
		
		self.footer_progress_label = customtkinter.CTkLabel(self.footer_frame, text="Stopped", width=150, font=customtkinter.CTkFont(size=16, weight="bold"), anchor="w")
		self.footer_progress_label.grid(row=0, column=0, padx=(20, 5), pady=header_padding_y)

		self.footer_progress_bar = customtkinter.CTkProgressBar(self.footer_frame)
		self.footer_progress_bar.grid(row=0, column=4, padx=(50, 100), pady=header_padding_y, sticky="ew")
		self.footer_progress_bar.set(0)

		play_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "play.png")), size=(20, 20))
		self.run_button = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=play_image, command=self.run)
		ctt(self.run_button, delay=0.1, alpha=0.95, message="Start Job")
		self.run_button.grid(row=0, column=7, padx=(20, 5), pady=header_padding_y)

		test_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "test.png")), size=(20, 20))
		self.test_button = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=test_image, command=self.run_test)
		ctt(self.test_button, delay=0.1, alpha=0.95, message="Start Testing")
		self.test_button.grid(row=0, column=8, padx=(5, 5), pady=header_padding_y)
		
		stop_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "stop.png")), size=(20, 20))
		self.stop_button = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=stop_image, command=self.stop)
		ctt(self.stop_button, delay=0.1, alpha=0.95, message="Stop")
		self.stop_button.grid(row=0, column=9, padx=(5, 20), pady=header_padding_y)

		download_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "down.png")), size=(20, 20))
		self.download_button = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=download_image, command=self.stop)
		ctt(self.download_button, delay=0.1, alpha=0.95, message="Download Results")
		self.download_button.grid(row=0, column=10, padx=(5, 20), pady=header_padding_y)

		self.refresh_step_view(0)
		self.after(1000, self.running_loop)

		NW(title="Welcome!", message="Welcome to COSU Manager!")

	def update_graph(self, value):
		self.vis_sidebar.refresh()
		GraphGenerator.generate_graphs(self)
			
	def save_project(self):
		filename = asksaveasfilename(filetypes=[("JSON", "*.json")], initialfile="config", defaultextension="json", title="Save Project")
		
		try:
			self.option_manager.set_path(filename)
			self.option_manager.save_project(filename)

			folder = self.option_manager.get_project_folder()
			if not os.path.exists(folder):
				os.makedirs(folder)

			if not os.path.exists(os.path.join(folder, "results")):
				os.makedirs(os.path.join(folder, "results"))
				
			self.save_button.configure(text="Saved!")
			self.after(3000, lambda: self.save_button.configure(text="Save"))
		except Exception as e:
			self.save_button.configure(text="Error!")
			print(e)
			self.after(3000, lambda: self.save_button.configure(text="Save"))
	
	def refresh_step_view(self, value):
		mode = self.option_manager.get_mode()
		self.service_url.configure(textvariable=self.option_manager.get("url"))
		self.graph_theme.configure(variable=self.option_manager.get("graph_theme"))

		# Limit graph selection based on mode
		self.graph_selector.configure(variable=self.option_manager.get("selected_graph"))
		graph_types = []
		if mode == "Optimization":
			graph_types = ["Best Cost Stacked", "Best Cost by Round", "Custom CSV", "Compare CSV"]
		elif mode == "Sensitivity Analysis":
			graph_types = ["Custom CSV", "Compare CSV"]
		elif mode == "Sampling: Halton" or mode == "Sampling: Random":
			graph_types = ["Sampling CSV", "Matrix Editor"]
		self.graph_selector.configure(values=graph_types)
		selected_graph = self.option_manager.get("selected_graph").get()
		if selected_graph not in graph_types:
			self.option_manager.get("selected_graph").set(graph_types[0])
		self.graph_selector.set(self.option_manager.get("selected_graph").get())

		if mode == "Optimization":
			self.test_button.configure(state="normal")
			self.test_button.configure(fg_color=get_color("CTkButton"))
			self.download_button.configure(state="normal")
			self.download_button.configure(fg_color=get_color("CTkButton"))
		else:
			self.test_button.configure(state="disabled")
			self.test_button.configure(fg_color="gray")
			self.download_button.configure(state="disabled")
			self.download_button.configure(fg_color="gray")

		self.vis_sidebar.clear()
		self.vis_sidebar.render()

		self.steps_frame.clear()
		self.steps_frame.render()

		self.static_param_frame.clear()
		self.static_param_frame.render()
		
		self.calib_param_frame.clear()
		self.calib_param_frame.render()

		self.service_param_frame.clear()
		self.service_param_frame.render()

		self.optimal_param_frame.clear()
		self.optimal_param_frame.render()

		self.project_name_label.configure(text=self.option_manager.get_project_data()["name"])
		self.project_path_label.configure(text=self.option_manager.get_project_data()["path"])

	def open_new_window(self):
		# Shell out and run ./main.py
		subprocess.Popen(["python3", "../mgpsogui.py"])
	
	def load_project(self):
		
		filename = askopenfilename(filetypes=[("JSON", "*.json")], title="Open Project", multiple=False)
		print(filename)
		
		try:
		
			self.option_manager.load_project(filename)
			self.option_manager.set_path(filename)

			folder = self.option_manager.get_project_folder()
			if not os.path.exists(folder):
				os.makedirs(folder)

			if not os.path.exists(os.path.join(folder, "results")):
				os.makedirs(os.path.join(folder, "results"))
			
			self.refresh_step_view(0)
			
			self.load_button.configure(text="Loaded!")
			self.after(3000, lambda: self.load_button.configure(text="Load"))

			self.load()
			
		except Exception as e:
			print(e)
			# Print stack trace 
			traceback.print_exc()
			

			self.load_button.configure(text="Error!")
			self.after(3000, lambda: self.load_button.configure(text="Load"))

	def change_appearance_mode_event(self, new_appearance_mode: str):
		customtkinter.set_appearance_mode(new_appearance_mode)

	def change_scaling_event(self, new_scaling: str):
		new_scaling_float = int(new_scaling.replace("%", "")) / 100
		customtkinter.set_widget_scaling(new_scaling_float)
		
	def change_scaling_event(self, new_scaling: str):
		new_scaling_float = int(new_scaling.replace("%", "")) / 100
		customtkinter.set_widget_scaling(new_scaling_float)
		
	def make_request(self):
		service_url = self.service_url.get()
		try:
			response = requests.get(service_url, timeout=5)
				
			response_json = json.loads(response.text)
			status = response.status_code
			
			self.option_manager.set_data("service_request_data", response_json)
			
			self.service_status.delete('0.0', tk.END)
			self.service_status.insert(text=str(status), index='0.0')
			self.service_name.delete('0.0', tk.END)
			self.service_name.insert(text=str(response_json["metainfo"]["name"]), index='0.0')
			self.service_description.delete('0.0', tk.END)
			self.service_description.insert(text=str(response_json["metainfo"]["description"]), index='0.0')
			self.service_details.delete('0.0', tk.END)
			self.service_details.insert(text=json.dumps(response_json, indent=4), index='0.0')

			self.refresh_button.configure(fg_color="green")
		except Exception as e:
			self.refresh_button.configure(fg_color="red")
			self.after(1000, lambda: self.refresh_button.configure(fg_color=get_color("CTkButton")))

		self.refresh_step_view(0)
		
	
	def load(self):
		# Make HTTP request to service_url and save the result to bounds.json
		
		self.refresh_button.configure(fg_color="gray")
		
		self.after(10, self.make_request)
		self.after(3000, lambda: self.refresh_button.configure(fg_color=get_color("CTkButton")))
		
	def run_test(self):
		self.testing = True
		self.run()	

	def run(self):
		
		data = self.option_manager.get_all_data()
		mode = self.option_manager.get_mode()
		data = data[mode]

		self.running_config = data
		
		#if self.testing:
		#	steps = metrics['steps']
		#	for step in steps:
		#		for param in step['param']:
		#			param['default_value'] = param['optimal_value']
		self.testing = False

		self.progress_data = None
		self.calibration_data = None

		#self.progress_bar.configure(mode="indeterminnate")
		#self.progress_bar.start()
		self.footer_progress_bar.configure(mode="indeterminnate")
		self.footer_progress_bar.start()
		
		self.data_x = [0]
		self.data_y = [0]
		
		#self.progress_message_middle.configure(text="Job starting...")
		self.footer_progress_label.configure(text="Starting...")
		
		self.textbox.insert("0.0", "Starting job...\n\n")
		self.textbox.insert("0.0", "Job Parameters:\n")
		self.textbox.insert("0.0", json.dumps(data, indent=4) + "\n\n")
		try:
			folder = self.option_manager.get_project_folder()
			
			self.train_process = Process(target=PSORunner.run_process, args=(stdout_queue, stderr_queue, results_queue, data, folder, mode))
			self.train_process.daemon = True
			self.train_process.start()
			self.after(1000, self.watch_loop)
			self.string_cache = ""
			self.data_cache = ""
			
		except Exception as e:
			self.textbox.insert("0.0", "An exception occurred!\n Exception: " + str(e) + "\n\n")
			self.textbox.insert("0.0", "Stack trace:\n")
			self.textbox.insert("0.0", traceback.format_exc())
			self.textbox.insert("0.0", "\n\n")
			self.textbox.insert("0.0", "Job failed!")
			#self.progress_message_left.configure(text="")
			#self.progress_message_middle.configure(text="Job failed! See error log below.")
			#self.progress_message_right.configure(text="")
			self.footer_progress_label.configure(text="Failed")
			#self.progress_bar.stop()
			self.footer_progress_bar.stop()
			#self.progress_bar.configure(mode="determinate")
			self.footer_progress_bar.configure(mode="determinate")
			#self.progress_bar.set(0)
			self.footer_progress_bar.set(0)
	
			
	def stop(self):
		print("Stopping...")
		self.train_process.terminate()
		
		folder = self.option_manager.get_project_folder()
			
		if not os.path.exists(folder):
			os.makedirs(folder)
		
		# Stop the process
		#if (os.path.exists(os.path.join(folder, 'output.txt'))):
		#	os.remove(os.path.join(folder, 'output.txt'))
		
		#if (os.path.exists(os.path.join(folder, 'error.txt'))):
		#	os.remove(os.path.join(folder, 'error.txt'))
			
		self.textbox.insert("0.0", "Job terminated!\n")
		#self.progress_bar.stop()
		self.footer_progress_bar.stop()
		#self.progress_bar.configure(mode="determinate")
		self.footer_progress_bar.configure(mode="determinate")
		#self.progress_bar.set(0)
		self.footer_progress_bar.set(0)
		#self.progress_message_left.configure(text="")
		#self.progress_message_middle.configure(text="Job stopped!")
		self.footer_progress_label.configure(text="Stopped")
		#self.progress_message_right.configure(text="")
		
	def running_loop(self):
		try:
			active_tab = self.tabview.get()
			if (active_tab == "Results"):
				GraphGenerator.generate_graphs(self)
		finally:
			self.after(1000, self.running_loop)

	def watch_loop(self):
		print("Watch loop running...")

		folder = self.option_manager.get_project_folder()
			
		if not os.path.exists(folder):
			os.makedirs(folder)
			
		while True:
			try:
				stdout_line = stdout_queue.get_nowait()
				
				print("# " + stdout_line, flush=True)
				with open(os.path.join(folder, 'output.txt'), 'a') as f:
					f.write(stdout_line)
			except Empty:
				break
			
		while True:
			try:
				stderr_line = stderr_queue.get_nowait()
				print("? " + stderr_line, flush=True)
				with open(os.path.join(folder, 'error.txt'), 'a') as f:
					f.write(stderr_line)
				
			except Empty:
				break

		while True:
			try:
				trace = results_queue.get_nowait()
				
				print("TRACE " + str(trace), flush=True)
	
				final_results = hp.extract_final_round_values(trace)
				print("Final results: " + str(final_results), flush=True)
	
				all_steps = self.option_manager.get_steps()
				data = final_results["data"]
				index = 0
				for step in data:
					for param in step.keys():
						value = step[param]
						target_step = all_steps[index]
						for target_param in target_step['parameter_objects']:
							if target_param['name'].get() == param:
								target_param['optimal_value'].set(value)
					index += 1
				print("Applied optimal values")
			except Empty:
				break
			except Exception as e:
				print("Some error happened when getting the trace!")
				print(e)
		
		# Write output to textbox
		if (os.path.exists(os.path.join(folder, 'output.txt'))):
			with open(os.path.join(folder, 'output.txt'), 'r') as f:
				lines = f.readlines()
				lines_string = "".join(lines)
				
				new_characters = lines_string.replace(self.string_cache, "")
				# Update the textbox with characters not in self.string_cache
				self.textbox.insert('0.0', new_characters)
				self.string_cache = lines_string
				print(new_characters, end="")
				
			try:
				# REPLACE THIS WITH READING THE STEP TRACE JSON for PSO
				with open(os.path.join(folder, "output.txt"), "r") as f:
					text = f.read()

				calibrated_params_pattern = r"calibrated params: ({.*?})"
				best_particle_values_pattern = r"best particle values: (\[.*?\])"
				progress_pattern = r"Progress -  best_round_cost:(.*?), rel_round_tol:(.*?), rtol:(.*?)\n"

				calibrated_params = re.findall(calibrated_params_pattern, text)
				best_particle_values = re.findall(best_particle_values_pattern, text)
				progress_values = re.findall(progress_pattern, text)

				for index, pp in enumerate(best_particle_values):
					pp = pp.strip()
					pp = pp.replace('[ ', '[')
					pp = pp.replace('  ', ',')
					pp = pp.replace(' ', ',')
					best_particle_values[index] = pp

				calibrated_params = [ast.literal_eval(i) for i in calibrated_params]
				best_particle_values = [ast.literal_eval(i) for i in best_particle_values]
				progress_values = [tuple(map(float, i)) for i in progress_values]
				self.calibration_data = calibrated_params
				
			except Exception as e:
				traceback.print_exc()
				print(e)
				
		# Parse data into interface
		mode = self.option_manager.get_mode()
		if mode == "Optimization":
			if (os.path.exists(os.path.join(folder, 'error.txt'))):
				data = hp.parse_pso_error(os.path.join(folder, 'error.txt'), len(self.option_manager.get_steps()))
				print(data)
				self.progress_data = data["data"]
				self.footer_progress_bar.stop()
				self.footer_progress_bar.configure(mode="determinate")
				self.footer_progress_bar.set(data["percent"]/100)
				self.footer_progress_label.configure(text="Round: " + str(data["round"]) + " - Step: " + str(data["step"]))
		else:
			if (os.path.exists(os.path.join(folder, 'output.txt'))):
				data = hp.parse_sampling_output(os.path.join(folder, 'output.txt'))
				print(data)
				self.footer_progress_bar.stop()
				self.footer_progress_bar.configure(mode="determinate")
				self.footer_progress_bar.set(data["percent"]/100)
				self.footer_progress_label.configure(text=str(round(data["percent"])) + "%")
				
				
				
		if self.train_process.is_alive():
			self.after(1000, self.watch_loop)
		else:
			#self.progress_bar.stop()
			self.footer_progress_bar.stop()
			#self.progress_bar.configure(mode="indeterminate")
			self.footer_progress_bar.configure(mode="indeterminate")
			#self.progress_bar.start()
			self.footer_progress_bar.start()
			#self.progress_message_left.configure(text="")
			#self.progress_message_middle.configure(text="Job finished!")
			#self.progress_message_right.configure(text="")
			self.textbox.insert("0.0", "\nJob finished!\n")

			# IF "./crash.txt" exists write it to the textbox
			if (os.path.exists(os.path.join(folder, 'crash.txt'))):
				with open(os.path.join(folder, 'crash.txt'), 'r') as f:
					lines = f.readlines()
					lines_string = "".join(lines)
					print("CRASH OCCURRED!")
					print(lines_string)
					self.textbox.insert('0.0', lines_string)
					self.textbox.insert('0.0', "CRASH OCCURRED!\n")
				# Delete the file
				os.remove(os.path.join(folder, 'crash.txt'))

def get_color(class_name, property = "fg_color"):
    return ThemeManager.theme[class_name][property]

def start():
	app = App()
	app.mainloop()

if __name__ == "__main__":
	app = App()
	app.mainloop()


from customtkinter import CTkScrollableFrame
from customtkinter import CTkFrame
from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
from customtkinter import CTkOptionMenu
from customtkinter import CTkImage
from .CustomFunctionEditorWindow import CustomFunctionEditorWindow as CFEW
import os
import PIL
from PIL import Image
import tkinter as tk

class FunctionsList(CTkFrame):
    def __init__(self, *args,
                 option_manager: None,
                 step_index: 0,
                 **kwargs):
        super().__init__(*args, **kwargs)
        

        self.option_manager = option_manager
        self.step_index = step_index

        self.edit_mode = False
        self.render()
    
    def refresh(self, op=None):
        self.clear()
        self.render()

    def clear(self):
        self.containerFrame.destroy()
        
    def toggle_edit_mode(self):
        self.clear()
        self.edit_mode = not self.edit_mode
        self.render()
        
    def render(self):
        
        self.containerFrame = CTkFrame(self, fg_color="transparent")
        self.containerFrame.grid(row=0, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
        self.containerFrame.grid_columnconfigure((0, 1), weight=1)
        
        row = 1
        
        funcs = self.option_manager.get_steps()[self.step_index]["objective_functions"]
        
        index = 0
        for func in funcs:
            
            CTkLabel(self.containerFrame, text="Name:").grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
            CTkLabel(self.containerFrame, text="Objective:").grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="nsew")
            CTkLabel(self.containerFrame, text="Weight:").grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="nsew")
            
            row += 1
            
            columns = ["absdiff",
                       "absdifflog",
                       "ave",
                       "bias",
                       "fhf",
                       "ioa",
                       "ioa2",
                       "kge",
                       "kge09",
                       "mns",
                       "mse",
                       "ns",
                       "ns2log",
                       "nslog1p",
                       "nslog2",
                       "pbias",
                       "pmcc",
                       "rmse",
                       "trmse",
                       "custom"]
            
            CTkEntry(self.containerFrame, textvariable=func["name"]).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="ew")
            CTkOptionMenu(self.containerFrame, values=columns, variable=func["objective_function"], command=self.refresh).grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="ew")
            CTkEntry(self.containerFrame, textvariable=func["weight"]).grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="ew")
            
            row += 1

            if func["objective_function"].get() == "custom":

                CTkLabel(self.containerFrame, text="Custom Function:").grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="nsew")
                #CTkLabel(self.containerFrame, text="Target:").grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="nsew")

                row += 1

                CTkEntry(self.containerFrame, textvariable=func["custom_function"]).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
                #CTkOptionMenu(self.containerFrame, values=["Positive Best", "Zero Best"], variable=func["custom_function_goal"]).grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="ew")

                def button_click_event(function_index):
                    dialog = CFEW(title="Edit Custom Function", step_index=self.step_index, function_index=function_index, option_manager=self.option_manager)


                open_window = lambda event=None, function_index=index: (button_click_event(function_index))
                expand_image = CTkImage(Image.open(os.path.join("./images", "expand.png")), size=(20, 20))
                button = CTkButton(self.containerFrame, width=30, text=None, image=expand_image, command=open_window)
                button.grid(row=row, column=3, padx=(5, 5), pady=(5, 5), sticky="new")


                row += 1
            
            CTkLabel(self.containerFrame, text="Observed and Simulated Data:").grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="nsew")
            
            row += 1
            
            CTkEntry(self.containerFrame, textvariable=func["data_observed"]).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="nsew")
            row += 1
            CTkEntry(self.containerFrame, textvariable=func["data_simulated"]).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="nsew")
            row += 1

            if self.edit_mode:
                remove_func = lambda index=index: (self.clear(), self.option_manager.remove_function(self.step_index, index), self.render())
                dupe_func = lambda index=index: (self.clear(), self.option_manager.dupe_function(self.step_index, index), self.render())
                CTkButton(self.containerFrame, text="Duplicate", command=dupe_func).grid(row=row, column=0, columnspan=1, padx=(5, 5), pady=(5, 5), sticky="ew")
                CTkButton(self.containerFrame, text="Remove", command=remove_func).grid(row=row, column=1, columnspan=1, padx=(5, 5), pady=(5, 5), sticky="ew")
                row += 1
                pass
            else:
                pass
            index += 1
            row += 1
            
        add_function = lambda: (self.clear(), self.option_manager.add_function(self.step_index), self.render())
        if len(funcs) > 0:
            if self.edit_mode:
                CTkButton(self.containerFrame, text="Exit", command=self.toggle_edit_mode).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="ew")
            else:
                CTkButton(self.containerFrame, text="Edit", command=self.toggle_edit_mode).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="ew")
            CTkButton(self.containerFrame, text="Add Objective Function", command=add_function).grid(row=row, column=1, columnspan=2, padx=(5, 5), pady=(5, 5), sticky="ew")
        else:
            CTkButton(self.containerFrame, text="Add Objective Function", command=add_function).grid(row=row, column=0, columnspan=2, padx=(5, 5), pady=(5, 5), sticky="ew")


import requests
import json
import os
import subprocess
import customtkinter
import platform

from ..General import ParameterView as pv

import tkinter as tk

def create_tab(self, tab):
    
    def make_request():
        service_url = self.service_url.get()
        response = requests.get(service_url)
            
        response_json = json.loads(response.text)
        status = response.status_code
        
        self.option_manager.set_data(response_json)
        self.tabview.configure(state="enabled")
        
        self.service_status.delete('0.0', tk.END)
        self.service_status.insert(text=str(status), index='0.0')
        self.service_name.delete('0.0', tk.END)
        self.service_name.insert(text=str(response_json["metainfo"]["name"]), index='0.0')
        self.service_description.delete('0.0', tk.END)
        self.service_description.insert(text=str(response_json["metainfo"]["description"]), index='0.0')
        self.service_details.delete('0.0', tk.END)
        self.service_details.insert(text=json.dumps(response_json, indent=4), index='0.0')
        
        self.load_parameters.configure(text="Loaded!")
    
    def load():
        # Make HTTP request to service_url and save the result to bounds.json
        self.load_parameters.configure(text="Loading...")
        self.after(10, make_request)
        self.after(3000, lambda: self.load_parameters.configure(text="Connect"))
        
    def start_cluster():
        process = subprocess.Popen(["minikube", "start"], stdout=subprocess.PIPE)
        output, error = process.communicate()
        if error:
            print(f"An error occurred: {error}")
            self.custer_details.insert("0.0", error.decode('unicode-escape') + "\n\n")
        else:
            print(f"Output: {output}")
            self.custer_details.insert("0.0", output.decode('unicode-escape') + "\n\n")
            
        process = subprocess.Popen(["minikube", "kubectl", "--", "create", "namespace", "csip"], stdout=subprocess.PIPE)
        output, error = process.communicate()
        if error:
            print(f"An error occurred: {error}")
            self.custer_details.insert("0.0", error.decode('unicode-escape') + "\n\n")
        else:
            print(f"Output: {output}")
            self.custer_details.insert("0.0", output.decode('unicode-escape') + "\n\n")
            
    def cluster_status():
        process = subprocess.Popen(["minikube", "status"], stdout=subprocess.PIPE)
        output, error = process.communicate()
        if error:
            print(f"An error occurred: {error}")
            self.custer_details.insert("0.0", error.decode('unicode-escape') + "\n\n")
        else:
            print(f"Output: {output}")
            self.custer_details.insert("0.0", output.decode('unicode-escape') + "\n\n")
            
                
    def open_terminal_and_run_cluster():
        full_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', 'start.yaml'))
        command = "minikube kubectl -- create -f " + full_path + " ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip"
        #command = "minikube kubectl -- create -f " + full_path + " ; until [[ $(minikube kubectl -- get pods -l app=pf8087-csu-csip-oms -n csip -o \\'jsonpath={..status.conditions[?(@.type==\\\"Ready\\\")].status}\\') == \\\"True\\\" ]]; do echo \\\"waiting for pod\\\" && sleep 1; done ; minikube service pf8087-csu-csip-oms -n csip"
        
        self.create_environment_button.configure(text="Running!")
        
        os.system("osascript -e 'tell app \"Terminal\" to do script \"" + command + "\"'")
        
        self.after(3000, lambda: self.create_environment_button.configure(text="Start Environment"))
        
        # Change the service URL to /csip-oms/m/ages/0.3.0
        self.service_url.cget("textvariable").set("PASTE_URL/csip-oms/m/ages/0.3.0")
        
    def deploy_cluster():
        full_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'start.yaml'))
        command = "minikube kubectl -- create -f " + full_path + " ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip"
        
        #minikube kubectl -- create -f ../start.yaml ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip
        
        # Get the full path of ../start.yamp
        full_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'start.yaml'))
        command = "minikube kubectl -- create -f " + full_path + " ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip"
        
        # Put that command in your clipboard with tinkter
        self.clipboard_clear()
        self.clipboard_append(command)
        
        # Change the button text breafly to "Copied!"
        self.create_environment_button.configure(text="Copied!")
        self.after(3000, lambda: self.create_environment_button.configure(text="Copy Command"))
        
        # Change the service URL to /csip-oms/m/ages/0.3.0
        self.service_url.cget("textvariable").set("PASTE_URL/csip-oms/m/ages/0.3.0")
        
    def stop_cluster():
        process = subprocess.Popen(["minikube", "delete"], stdout=subprocess.PIPE)
        output, error = process.communicate()
        if error:
            print(f"An error occurred: {error}")
            self.custer_details.insert("0.0", error.decode('latin_1') + "\n\n")
        else:
            print(f"Output: {output}")
            self.custer_details.insert("0.0", output.decode('latin_1') + "\n\n")
    
    
    tab.grid_columnconfigure(0, weight=1)
    tab.grid_columnconfigure(1, weight=1)
    tab.grid_columnconfigure(2, weight=1)
    tab.grid_rowconfigure(0, weight=1)
    
    """
    self.top_bar_container = customtkinter.CTkFrame(tab)
    self.top_bar_container.grid(row=0, column=0, columnspan=2, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.top_bar_container.grid_columnconfigure(1, weight=1)
    
    cl = customtkinter.CTkLabel(self.top_bar_container, text="Service URL:", anchor="w")
    cl.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="ew")
    
    self.service_url = customtkinter.CTkEntry(self.top_bar_container, textvariable=self.option_manager.get_arguments()['url'])
    self.service_url.grid(row=0, column=1, columnspan=1, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    self.load_parameters = customtkinter.CTkButton(self.top_bar_container, text="Connect", command=load)
    self.load_parameters.grid(row=0, column=2, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    """

    self.service_param_frame = pv.ParameterView(tab, option_manager=self.option_manager, list_name="service_parameters", label_text="Service Parameters")
    self.service_param_frame.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.service_param_frame.grid_columnconfigure(0, weight=1)
    self.service_param_frame.grid_rowconfigure(0, weight=1)

            
    self.service_editor = customtkinter.CTkScrollableFrame(tab, label_text="Service Status")
    self.service_editor.grid(row=0, column=1, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.service_editor.grid_columnconfigure(0, weight=1)
    self.service_editor.grid_rowconfigure(0, weight=1)
    
    cl = customtkinter.CTkLabel(self.service_editor, text="Service Status:", anchor="w")
    cl.grid(row=1, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.service_status = customtkinter.CTkTextbox(self.service_editor, height=32)
    self.service_status.grid(row=2, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    cl = customtkinter.CTkLabel(self.service_editor, text="Service Name:", anchor="w")
    cl.grid(row=3, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.service_name = customtkinter.CTkTextbox(self.service_editor, height=32)
    self.service_name.grid(row=4, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    cl = customtkinter.CTkLabel(self.service_editor, text="Service Description:", anchor="w")
    cl.grid(row=5, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.service_description = customtkinter.CTkTextbox(self.service_editor, height=32)
    self.service_description.grid(row=6, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    cl = customtkinter.CTkLabel(self.service_editor, text="Service Details:", anchor="w")
    cl.grid(row=7, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.service_details = customtkinter.CTkTextbox(self.service_editor, height=480)
    self.service_details.grid(row=8, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.environment_editor = customtkinter.CTkScrollableFrame(tab, label_text="Minikube Environment Editor")
    self.environment_editor.grid(row=0, column=2, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.environment_editor.grid_columnconfigure(0, weight=1)
    self.environment_editor.grid_rowconfigure(0, weight=1)
    
    cl = customtkinter.CTkLabel(self.environment_editor, text="Cluster Preset:", anchor="w")
    cl.grid(row=0, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.preset_selector = customtkinter.CTkOptionMenu(self.environment_editor, values=["Local", "Other"])
    self.preset_selector.grid(row=1, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    cl = customtkinter.CTkLabel(self.environment_editor, text="Controls:", anchor="w")
    cl.grid(row=2, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.start_cluster_button = customtkinter.CTkButton(self.environment_editor, text="Start Minikube", command=start_cluster)
    self.start_cluster_button.grid(row=3, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")
    
    # Check if the os is MacOS
    if (platform.system() == "Darwin"):
        self.create_environment_button = customtkinter.CTkButton(self.environment_editor, text="Create Environment", command=open_terminal_and_run_cluster)
    else:
        self.create_environment_button = customtkinter.CTkButton(self.environment_editor, text="Copy Command", command=deploy_cluster)
        
    self.create_environment_button.grid(row=4, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")
    
    self.get_status = customtkinter.CTkButton(self.environment_editor, text="Minikube Status", command=cluster_status)
    self.get_status.grid(row=5, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")
    
    self.destroy_environment_button = customtkinter.CTkButton(self.environment_editor, text="Stop Minikube", command=stop_cluster)
    self.destroy_environment_button.grid(row=6, column=0, padx=(20, 20), pady=(5,10), sticky="ew")
    
    cl = customtkinter.CTkLabel(self.environment_editor, text="Minikube Status:", anchor="w")
    cl.grid(row=7, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")
    
    self.custer_details = customtkinter.CTkTextbox(self.environment_editor, height=480)
    self.custer_details.grid(row=8, column=0, padx=(20, 20), pady=(5, 5), sticky="ew")



import customtkinter
import json
import os
from multiprocessing import Process
import traceback
import re
import ast
import pandas as pd
import numpy as np
import os
import platform
import subprocess

from ...util import PSORunner
from ...util import GraphGenerator
from ...util.CTkToolTip import CTkToolTip as ctt
import customtkinter

from . import OptimalParameterView as opv

def create_tab(self, tab):
    
    # URL
    tab.grid_columnconfigure(0, weight=1)
    tab.grid_columnconfigure(1, weight=5)
    tab.grid_columnconfigure(2, weight=1)
    tab.grid_rowconfigure(0, weight=200)

    #self.progress_container = customtkinter.CTkFrame(tab)
    #self.progress_container.grid_columnconfigure(0, weight=1)
    #self.progress_container.grid_columnconfigure(1, weight=1)
    #self.progress_container.grid_columnconfigure(2, weight=1)
    #self.progress_container.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    # Add progress bar to progress container
    #self.progress_message_left = customtkinter.CTkLabel(self.progress_container, text="")
    #self.progress_message_left.grid(row=0, column=0, padx=(10, 10), pady=(10, 10), sticky="w")
    
    #self.progress_message_middle = customtkinter.CTkLabel(self.progress_container, text="Calibration not running...")
    #self.progress_message_middle.grid(row=0, column=1, padx=(10, 10), pady=(10, 10), sticky="ew")
    
    #self.progress_message_right = customtkinter.CTkLabel(self.progress_container, text="")
    #self.progress_message_right.grid(row=0, column=2, padx=(10, 10), pady=(10, 10), sticky="e")
    
    self.optimal_param_frame = opv.OptimalParameterView(tab, option_manager=self.option_manager, label_text="Optimal Parameters")
    self.optimal_param_frame.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.optimal_param_frame.grid_columnconfigure(0, weight=1)
    self.optimal_param_frame.grid_rowconfigure(0, weight=1)

    self.textbox = customtkinter.CTkTextbox(tab)
    self.textbox.grid(row=0, column=1, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.textbox.insert("0.0", "Welcome to the CSIP PSO Calibration Tool!\n\nUse the Setup tab to define steps and calibration parameters. Use this tab to view logs and manage your Project. Once finished, use the Results tab to generate figures and graphs.")
        
    self.project_editor = customtkinter.CTkScrollableFrame(tab, label_text="Project Editor")
    self.project_editor.grid(row=0, column=2, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.project_editor.grid_columnconfigure(0, weight=1)
    self.project_editor.grid_rowconfigure(0, weight=1)

    customtkinter.CTkLabel(self.project_editor, text="Copy Data to Current Mode:").grid(row=0, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    def copy_sense():
        self.option_manager.copy_list("Sensitivity Analysis")
        self.refresh_step_view(0)

    def copy_halton():
        self.option_manager.copy_list("Sampling: Halton")
        self.refresh_step_view(0)
    
    def copy_random():
        self.option_manager.copy_list("Sampling: Random")
        self.refresh_step_view(0)

    def copy_optim():
        self.option_manager.copy_list("Optimization")
        self.refresh_step_view(0)

    def open_project_folder():
        # Open the file in the default program
        folder = self.option_manager.get_project_folder()
        
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        if platform.system() == "Windows":
            os.startfile(folder)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", folder])
        else:
            subprocess.Popen(["xdg-open", folder])

    self.halton_copy = customtkinter.CTkButton(self.project_editor, text="From Sampling: Halton", command=copy_halton)
    self.halton_copy.grid(row=1, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    self.random_copy = customtkinter.CTkButton(self.project_editor, text="From Sampling: Random", command=copy_random)
    self.random_copy.grid(row=2, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    self.analysis_copy = customtkinter.CTkButton(self.project_editor, text="From Sensitivity Analysis", command=copy_sense)
    self.analysis_copy.grid(row=3, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    self.optimization_copy = customtkinter.CTkButton(self.project_editor, text="From Optimization", command=copy_optim)
    self.optimization_copy.grid(row=4, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    customtkinter.CTkLabel(self.project_editor, text="Project Management:").grid(row=5, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    self.open_folder = customtkinter.CTkButton(self.project_editor, text="Open Project Folder", command=open_project_folder)
    self.open_folder.grid(row=6, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    customtkinter.CTkLabel(self.project_editor, text="Project Name:").grid(row=7, column=0, padx=(20, 20), pady=(10, 5), sticky="ew")
    self.project_name_label = customtkinter.CTkLabel(self.project_editor, text=self.option_manager.get_project_data()["name"])
    self.project_name_label.grid(row=8, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")

    customtkinter.CTkLabel(self.project_editor, text="Project Path:").grid(row=9, column=0, padx=(20, 20), pady=(10, 5), sticky="ew")
    self.project_path_label = customtkinter.CTkLabel(self.project_editor, text=self.option_manager.get_project_data()["path"])
    self.project_path_label.grid(row=10, column=0, padx=(20, 20), pady=(5, 10), sticky="ew")


import customtkinter
import tkinter as tk
from PIL import Image, ImageTk
import os
import platform
import subprocess
import shutil

from . import SideBar

def create_tab(self, tab):
    
    def open_graph_in_browser():
        # Open the file in the default program
        folder = self.option_manager.get_project_folder()
        
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        file_path = os.path.join(folder, self.selected_graph_name + ".html")
        
        if platform.system() == "Windows":
            os.startfile(file_path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", file_path])
        else:
            subprocess.Popen(["xdg-open", file_path])

    def export_graph():
        
        folder = self.option_manager.get_project_folder()
        
        if not os.path.exists(folder):
            os.makedirs(folder)

        html_file = os.path.join(folder, self.selected_graph_name + ".html")
        json_file = os.path.join(folder, self.selected_graph_name + ".json")
        png_file = os.path.join(folder, self.selected_graph_name + ".png")
        pdf_file = os.path.join(folder, self.selected_graph_name + ".pdf")

        # Make a directory call package
        if not os.path.exists(os.path.join(folder, self.selected_graph_name + "_package")):
            os.makedirs(os.path.join(folder, self.selected_graph_name + "_package"))

        # Check if html_file exists and copy it to the package directory
        if os.path.exists(html_file):
            with open(html_file, 'r') as file:
                data = file.read()
                with open(os.path.join(folder, self.selected_graph_name + "_package", self.selected_graph_name + ".html"), 'w') as new_file:
                    new_file.write(data)

        # Check if json_file exists and copy it to the package directory
        if os.path.exists(json_file):
            with open(json_file, 'r') as file:
                data = file.read()
                with open(os.path.join(folder, self.selected_graph_name + "_package", self.selected_graph_name + ".json"), 'w') as new_file:
                    new_file.write(data)
        
        # Check if png_file exists and copy it to the package directory
        if os.path.exists(png_file):
            with open(png_file, 'rb') as file:
                data = file.read()
                with open(os.path.join(folder, self.selected_graph_name + "_package", self.selected_graph_name + ".png"), 'wb') as new_file:
                    new_file.write(data)

        # Check if pdf_file exists and copy it to the package directory
        if os.path.exists(pdf_file):
            with open(pdf_file, 'rb') as file:
                data = file.read()
                with open(os.path.join(folder, self.selected_graph_name + "_package", self.selected_graph_name + ".pdf"), 'wb') as new_file:
                    new_file.write(data)
        
        # Zip the package directory
        shutil.make_archive(os.path.join(folder, self.selected_graph_name + "_package"), 'zip', os.path.join(folder, self.selected_graph_name + "_package"))

        # Open the directory containing the package
        if platform.system() == "Windows":
            os.startfile(os.path.join(folder, self.selected_graph_name + "_package"))
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", os.path.join(folder, self.selected_graph_name + "_package")])
        else:
            subprocess.Popen(["xdg-open", os.path.join(folder, self.selected_graph_name + "_package")])

    def _resize_image(event):
        self.graph_label.update_idletasks()
        new_width = self.graph_label.winfo_width()
        new_height = self.graph_label.winfo_height()
        
        alt_width = new_height * 1.77778
        alt_height = new_width / 1.77778
        
        if (new_width < new_height):
            new_height = alt_height
        else:
            new_width = alt_width
        
        self.image_width = new_width
        self.image_height = new_height
        
        self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(new_width, new_height))
        self.graph_label.configure(image=self.graph_image)
        self.graph_label.update_idletasks()
    
    
    tab.grid_columnconfigure(0, weight=2)
    tab.grid_columnconfigure(1, weight=8)
    tab.grid_rowconfigure(0, weight=1)
    
    self.graph_sidebar = customtkinter.CTkFrame(tab)
    self.graph_sidebar.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.graph_container = customtkinter.CTkFrame(tab)
    self.graph_container.grid(row=0, column=1, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    self.graph_sidebar.grid_columnconfigure(0, weight=1)
    self.graph_sidebar.grid_rowconfigure(1, weight=1)
    self.graph_container.grid_columnconfigure(0, weight=1)
    self.graph_container.grid_rowconfigure(0, weight=1)
    
    graph_types = []
    self.graph_selector = customtkinter.CTkOptionMenu(self.graph_sidebar, values=graph_types, variable=self.option_manager.get("selected_graph"), command=self.update_graph)
    self.graph_selector.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    # Create SideBar
    self.vis_sidebar = SideBar.SideBar(self.graph_sidebar, option_manager=self.option_manager, home_page=self, fg_color="transparent")
    self.vis_sidebar.grid(row=1, column=0, rowspan=6, padx=(0, 0), pady=(0, 0), sticky="nsew")
    

    self.graph_theme = customtkinter.CTkOptionMenu(self.graph_sidebar, values=["Dark", "Light", "Publication"], variable=self.option_manager.get("graph_theme"), command=self.update_graph)
    self.graph_theme.grid(row=7, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")

    self.graph_export = customtkinter.CTkButton(self.graph_sidebar, text="Export", command=export_graph)
    self.graph_export.grid(row=8, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")

    # Add a button to call open_graph_in_browser
    self.graph_button = customtkinter.CTkButton(self.graph_sidebar, text="Preview", command=open_graph_in_browser)
    self.graph_button.grid(row=9, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    self.graph_image_obj = Image.open(os.path.join("./images", "refresh_hd.png"))
    self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(1280, 720))
    self.graph_label = customtkinter.CTkLabel(self.graph_container, text=None, image=self.graph_image)
    self.graph_label.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
